/*  miscc -- a compiler for misc
 *    Semantic analyzer.
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Analyzer.java,v 1.1 2004/03/01 05:34:24 schinz Exp $
 */

package miscc;

import miscc.Tree.*;


/** This class implements the semantic analyzer of miscc.
 */
class Analyzer implements Visitor {

    Scope globalScope = new Scope();

    /** The current scope.
     */
    Scope scope = globalScope;

    {
        // enter a predefined function into the top-level scope
        globalScope.enter(
            new Symbol(0,
                "printInt",
                new Type.FunType(new Type[]{Type.INT_TYPE}, Type.UNIT_TYPE),
                false));
        globalScope.enter(
            new Symbol(0,
                "printChar",
                new Type.FunType(new Type[]{Type.INT_TYPE}, Type.UNIT_TYPE),
                false));
        globalScope.enter(
            new Symbol(0,
                "readInt",
                new Type.FunType(new Type[]{}, Type.INT_TYPE),
                false));
        globalScope.enter(
            new Symbol(0,
                "readChar",
                new Type.FunType(new Type[]{}, Type.INT_TYPE),
                false));
    }

    /** Analyze the tree with a given scope
     */
    Type analyze(Tree tree, Scope scope) {
        Scope backup = this.scope;
        this.scope = scope;
        Type type = analyze(tree);
        this.scope = backup;
        return type;
    }

    /** Analyze the tree in the current scope
     */
    Type analyze(Tree tree) {
        tree.apply(this);
        return tree.type;
    }

    /** Analyze several trees with a given scope.
     */
    private Type[] analyze(Tree[] trees, Scope scope) {
        Type[] types = new Type[trees.length];
        for (int i = 0; i < trees.length; i++)
            types[i] = analyze(trees[i], scope);
        return types;
    }

    /** Analyze several trees in the current scope.
     */
    private Type[] analyze(Tree[] trees) {
        return analyze(trees, scope);
    }

    /** Generates an error if the expected type is not equal
     *  to the found type.
     */
    private void checkSameType(int pos, Type found, Type expected) {
        if (!found.isSameType(expected))
            error(pos, found, expected);
    }

    /** Generates an error if the expected type is not a super type
     *  to the found type.
     */
    private void checkSubType(int pos, Type found, Type expected){
        if (!found.isSubType(expected))
            error(pos, found, expected);
    }

    private void error(int pos, Type found, Type expected) {
        Report.error(pos, "expected type: " + expected + "\n" +
                          "actual type  : " + found);
    }

    //// IMPLEMENTATION OF THE VISITOR INTERFACE ////

    public void caseBad(Bad tree) {
        tree.type = Type.BAD_TYPE;
    }

    public void caseProgram(Program tree) {
        Scope local = new Scope(scope);
        analyze(tree.defs, local);
        analyze(tree.body, local);
        tree.type = Type.NO_TYPE;
    }

    public void caseBody(Body tree) {
        Scope local = new Scope(scope);
        analyze(tree.stats, local);
        tree.type = analyze(tree.expr, local);
    }

    public void caseFunDef(FunDef tree) {
        Scope local = new Scope(scope);
        Type[] argTypes = analyze(tree.args, local);
        Type resType = analyze(tree.restype);
        tree.sym = new Symbol(tree.pos, tree.name,
                        new Type.FunType(argTypes, resType), false);
        scope.enter(tree.sym);
        checkSubType(tree.pos, analyze(tree.body, local), resType);
        tree.type = Type.NO_TYPE;
    }

    public void caseVarDef(VarDef tree) {
        analyze(tree.init);
        analyze(tree.var);
        checkSubType(tree.pos, tree.init.type, tree.var.type);
        tree.type = Type.NO_TYPE;
    }

    public void caseFormal(Formal tree) {
        tree.sym = new Symbol(tree.pos, tree.name,
                        analyze(tree.vartype), true);
        scope.enter(tree.sym);
        tree.type = tree.sym.type;
    }

    public void caseBasicType(BasicType tree) {
        switch (tree.tag) {
            case Tree.UNIT:
                tree.type = Type.UNIT_TYPE;
                break;
            case Tree.ANY:
                tree.type = Type.ANY_TYPE;
                break;
            case Tree.INT:
                tree.type = Type.INT_TYPE;
                break;
            default:
                throw new Error();
        }
    }

    public void caseListType(ListType tree) {
        tree.type = new Type.ListType(analyze(tree.elemtype));
    }

    public void caseFunType(FunType tree) {
        tree.type = new Type.FunType(
                        analyze(tree.argtypes), analyze(tree.restype));
    }

    public void caseExec(Exec tree) {
        analyze(tree.expr);
        tree.type = Type.NO_TYPE;
    }

    public void caseWhile(While tree) {
        checkSameType(tree.pos, analyze(tree.cond), Type.INT_TYPE);
        analyze(tree.body);
        tree.type = Type.NO_TYPE;
    }

    public void caseIf(If tree) {
        checkSameType(tree.pos, analyze(tree.cond), Type.INT_TYPE);
        tree.type = analyze(tree.thenp).max(analyze(tree.elsep));
        if (tree.type == Type.BAD_TYPE)
            Report.error(tree.pos,
                "incompatible branches of 'if' statement\n" +
                "type of then-part: " + tree.thenp.type + "\n" +
                "type of else-part: " + tree.elsep.type);
    }

    public void caseAssign(Assign tree) {
        tree.sym = scope.lookup(tree.name);
        if (tree.sym == null) {
            Report.error(tree.pos, "the variable " + tree.name +
                                   " is unknown.");
            tree.sym = new Symbol(tree.pos, tree.name, Type.BAD_TYPE, true);
        }
        if (!tree.sym.isVariable)
            Report.error(tree.pos, tree.name +
                                   " doesn't refer to a variable.");
        checkSubType(tree.pos, analyze(tree.expr), tree.sym.type);
        tree.type = tree.expr.type;
    }

    public void caseIdent(Ident tree) {
        tree.sym = scope.lookup(tree.name);
        if (tree.sym == null) {
            Report.error(tree.pos, "the variable/function " +
                                   tree.name + " is unknown.");
            tree.sym = new Symbol(tree.pos, tree.name, Type.BAD_TYPE, true);
        }
        tree.type = tree.sym.type;
    }

    public void caseUnitLit(UnitLit tree) {
        tree.type = Type.UNIT_TYPE;
    }

    public void caseIntLit(IntLit tree) {
        tree.type = Type.INT_TYPE;
    }

    public void caseNilLit(NilLit tree) {
        tree.type = new Type.ListType(Type.BOTTOM_TYPE);
    }

    public void caseOperation(Operation tree) {
        if (tree.op == Tree.CONS) {
            tree.type = analyze(tree.right).max(
                            new Type.ListType(analyze(tree.left)));
            if (tree.type == Type.BAD_TYPE)
                Report.error(tree.pos, "head doesn't fit to tail");
        } else {
            checkSameType(tree.left.pos, analyze(tree.left), Type.INT_TYPE);
            if (tree.right != null)
                checkSameType(tree.right.pos, analyze(tree.right), Type.INT_TYPE);
            tree.type = Type.INT_TYPE;
        }
    }

    public void caseListop(Listop tree) {
        if (tree.args.length != 1)
            Report.error(tree.pos, "wrong number of arguments");
        switch (tree.op) {
            case Tree.HEAD:
                if (analyze(tree.args[0]) instanceof Type.ListType)
                    tree.type = ((Type.ListType)tree.args[0].type).elemType;
                else {
                    Report.error(tree.args[0].pos, "expected list type");
                    tree.type = Type.BAD_TYPE;
                }
                break;
            case Tree.TAIL:
                checkSubType(tree.args[0].pos, analyze(tree.args[0]),
                             new Type.ListType(Type.ANY_TYPE));
                tree.type = tree.args[0].type;
                break;
            case Tree.ISEMPTY:
                checkSubType(tree.args[0].pos,
                             analyze(tree.args[0]),
                             new Type.ListType(Type.ANY_TYPE));
                tree.type = Type.INT_TYPE;
                break;
            default :
                throw new Error();
        }
    }

    public void caseApply(Apply tree) {
        checkSubType(tree.pos, analyze(tree.fun),
            new Type.FunType(analyze(tree.args), Type.ANY_TYPE));
        tree.type = (tree.fun.type instanceof Type.FunType) ?
            ((Type.FunType)tree.fun.type).resType : Type.BAD_TYPE;
    }
}
