/*  miscc -- a compiler for misc
 *    Support for code generation
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Code.java,v 1.2 2004/03/22 16:09:20 schinz Exp $
 */

package miscc;

import java.io.PrintWriter;
import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.BitSet;

public class Code implements RISC {
    //########################################################################
    // Constants

    /** Index of first and last general-use register. */
    protected final int regMin, regMax;

    //########################################################################
    // Instance variables

    /** Allocated registers */
    private BitSet usedRegisters = new BitSet();

    /** Already emitted instructions */
    private final List code;

    /** Stack frame size */
    private int frameSize;

    //########################################################################
    // Constructor

    protected Code(int regMin, int regMax) {
        code = new ArrayList();
        this.regMin = regMin;
        this.regMax = regMax;
        usedRegisters = new BitSet(32);
        frameSize = 0;
    }

    public Code() {
        this(1, 29);
    }

    //########################################################################
    // Methods - register allocation

    /** Allocate a register. */
    public int getRegister() {
        return getRegister(usedRegisters.nextClearBit(regMin));
    }

    /**
     * Allocate the given register, which has to be free (otherwise an
     * exception is thrown).
     */
    public int getRegister(int r) {
        assert !usedRegisters.get(r)
            : "attempt to allocate non-free register R" + r;
        assert (r >= regMin) && (r <= regMax)
            : "attempt to allocate invalid register R" + r;
        usedRegisters.set(r);
        return r;
    }

    /** Free a register. */
    public void freeRegister(int r) {
        assert usedRegisters.get(r)
            : "attempt to free non-allocated register R" + r;
        usedRegisters.clear(r);
    }

    /**
     * Return the set of used registers as an array. If element at
     * index I is true, this means that register R[i] is in use.
     */
    public BitSet usedRegisters() {
        return (BitSet)usedRegisters.clone();
    }

    public int freeRegistersCount() {
        return (regMax - regMin + 1) - usedRegistersCount();
    }

    /** Return the number of used registers. */
    public int usedRegistersCount() {
        return usedRegisters.cardinality();
    }

    //########################################################################
    // Methods - address handling

    /** Return address of next instruction. */
    public int pc() {
        return WORD_SIZE * code.size();
    }

    //########################################################################
    // Methods - frame handling

    /** Increment frame size by given value. */
    public void incFrameSize(int bytes) {
        frameSize += bytes;
    }

    /** Decrement frame size by given value. */
    public void decFrameSize(int bytes) {
        frameSize -= bytes;
        assert frameSize >= 0;
    }

    /** Return current frame size. */
    public int getFrameSize() {
        return frameSize;
    }

    //########################################################################
    // Methods - code emitting

    public void emit(int opcode) {
        emit(opcode, Integer.MIN_VALUE);
    }

    public void emit(int opcode, int c) {
        emit(opcode, Integer.MIN_VALUE, c);
    }

    public void emit(int opcode, Label l) {
        emit(opcode, Integer.MIN_VALUE, l);
    }

    public void emit(int opcode, int a, int c) {
        emit(opcode, a, Integer.MIN_VALUE, c);
    }

    public void emit(int opcode, int a, Label l) {
        if (l.isAnchored())
            emit(opcode, a, (l.getAnchor() - pc()) / WORD_SIZE);
        else {
            l.recordInstructionToPatch(pc());
            emit(opcode, a, Integer.MIN_VALUE, Integer.MIN_VALUE);
        }
    }

    public void emit(int opcode, int a, int b, int c) {
        code.add(new Instruction(opcode, a, b, c));
    }

    //########################################################################
    // Methods - program printing.

    /** Print program. */
    public void write(PrintWriter out) {
        for (int i = 0; i < code.size(); i++) {
            String label = Integer.toString(WORD_SIZE * i);
            while (label.length() < 4) label = '0' + label;
            out.print("/* " + label + " */ ");
            out.print(code.get(i));
            out.println();
        }
        out.flush();
    }

    //########################################################################
    // Labels

    public Label getLabel() {
        return new Label();
    }

    public void anchorLabel(Label l) {
        l.setAnchor(pc());
    }

    public class Label {
        private ArrayList/*<Integer>*/ toPatch = new ArrayList();
        private int pc = -1;

        private Label() {}
        private void setAnchor(int pc) {
            this.pc = pc;
            Iterator toPatchIt = toPatch.iterator();
            while (toPatchIt.hasNext()) {
                int instrPC = ((Integer)toPatchIt.next()).intValue();
                ((Instruction)code.get(instrPC / WORD_SIZE)).c =
                    (pc - instrPC) / WORD_SIZE;
            }
        }
        private int getAnchor() { return pc; }
        private boolean isAnchored() { return pc >= 0; }
        private void recordInstructionToPatch(int pc) {
            toPatch.add(new Integer(pc));
        }
    }

    //########################################################################
    // Instructions

    private static class Instruction {

        public final int opcode;
        public final int a;
        public final int b;
        public int c;

        public Instruction(int opcode, int a, int b, int c) {
            this.opcode = opcode;
            this.a = a;
            this.b = b;
            this.c = c;
        }

        public String toString() {
            StringBuffer buffer = new StringBuffer();
            buffer.append(mnemonics[opcode]);
            if (a != Integer.MIN_VALUE) buffer.append(' ').append(a);
            if (b != Integer.MIN_VALUE) buffer.append(' ').append(b);
            if (c != Integer.MIN_VALUE) buffer.append(' ').append(c);
            return buffer.toString();
        }
    }

    //########################################################################
}
