/*  miscc -- a compiler for misc
 *    Printer to graphviz's "dot" format.
 *
 *    Michel Schinz
 *    $Id: DotPrinter.java,v 1.2 2004/03/23 09:38:48 schinz Exp $
 */

package miscc;

import miscc.Tree.*;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Iterator;
import java.io.PrintStream;

/**
 * Generate a GraphViz (http://www.graphviz.org/) graph corresponding
 * to the tree.
 */

class DotPrinter implements Visitor {
    private Node self = null;

    public void print(Tree tree) {
        PrintStream out = System.out;

        nodeFor(tree);
        out.println("digraph misc_program {");

        Iterator nodeIt = nodes.values().iterator();
        while (nodeIt.hasNext()) {
            Node node = (Node)nodeIt.next();
            out.println(node + " ["
                        + " label = \"" + node.label + "\""
                        + ", style=filled"
                        + ", fillcolor = \"" + node.color + "\""
                        + " ];");
        }

        Iterator edgeIt = edges.iterator();
        while (edgeIt.hasNext()) {
            Edge edge = (Edge)edgeIt.next();
            assert edge.from != null;
            out.println(edge.from + " -> " + edge.to
                        + "[ label = \"" + edge.label + "\" ];");
        }

        out.println("}");
    }

    public void caseBad(Bad tree) {
        self = addNode(tree, "BAD");
    }

    public void caseProgram(Program tree) {
        self = addNode(tree, "Program");
        child(self, tree.defs);
        child(self, tree.body);
    }

    public void caseBody(Body tree) {
        self = addNode(tree, "Body");
        child(self, tree.stats);
        child(self, tree.expr);
    }

    public void caseFunDef(FunDef tree) {
        self = addNode(tree, "FunDef");
        child(self, tree.restype);
        child(self, tree.args);
        child(self, tree.body);
    }

    public void caseFormal(Formal tree) {
        self = addNode(tree, "Formal(" + tree.name + ")");
        child(self, tree.vartype);
    }

    public void caseBasicType(BasicType tree) {
        switch (tree.tag) {
        case Tree.UNIT: self = addNode(tree, "Unit"); break;
        case Tree.ANY: self = addNode(tree, "Any"); break;
        case Tree.INT: self = addNode(tree, "Int"); break;
        default: throw new Error();
        }
    }

    public void caseListType(ListType tree) {
        self = addNode(tree, "ListType");
        child(self, tree.elemtype);
    }

    public void caseFunType(FunType tree) {
        self = addNode(tree, "FunType");
        child(self, tree.argtypes);
        child(self, tree.restype);
    }

    public void caseVarDef(VarDef tree) {
        self = addNode(tree, "VarDef");
        child(self, tree.var);
        child(self, tree.init);
    }

    public void caseWhile(While tree) {
        self = addNode(tree, "While");
        child(self, tree.cond);
        child(self, tree.body);
    }

    public void caseExec(Exec tree) {
        self = addNode(tree, "Exec");
        child(self, tree.expr);
    }

    public void caseIf(If tree) {
        self = addNode(tree, "If");
        child(self, tree.cond);
        child(self, tree.thenp);
        child(self, tree.elsep);
    }

    public void caseAssign(Assign tree) {
        self = addNode(tree, "Assign(" + tree.name + ")");
        child(self, tree.expr);
    }

    public void caseOperation(Operation tree) {
        String opName;
        switch (tree.op) {
        case Tree.ADD: opName = "+"; break;
        case Tree.SUB: opName = "-"; break;
        case Tree.MUL: opName = "*"; break;
        case Tree.DIV: opName = "/"; break;
        case Tree.MOD: opName = "%"; break;
        case Tree.EQ: opName = "=="; break;
        case Tree.NE: opName = "!="; break;
        case Tree.LT: opName = "<"; break;
        case Tree.LE: opName = "<="; break;
        case Tree.GT: opName = ">"; break;
        case Tree.GE: opName = ">="; break;
        case Tree.NOT: opName = "!"; break;
        case Tree.AND: opName = "&&"; break;
        case Tree.CONS: opName = "::"; break;
        default: throw new Error("unknown op: " + tree.op);
        }
        self = addNode(tree, "Operation(" + opName + ")");
        child(self, tree.left, "l");
        if (tree.right != null)
            child(self, tree.right, "r");
    }

    public void caseListop(Listop tree) {
        String opName;
        switch (tree.op) {
        case Tree.HEAD: opName = "head"; break;
        case Tree.TAIL: opName = "tail"; break;
        case Tree.ISEMPTY: opName = "isEmpty"; break;
        default: throw new Error("unknown op: " + tree.op);
        }
        self = addNode(tree, "Listop(" + opName + ")");
        child(self, tree.args[0]);
    }

    public void caseIdent(Ident tree) {
        self = addNode(tree, "Ident(" + tree.name + ")");
    }

    public void caseUnitLit(UnitLit tree) {
        self = addNode(tree, "UnitLit");
    }

    public void caseNilLit(NilLit tree) {
        self = addNode(tree, "NilLit");
    }

    public void caseIntLit(IntLit tree) {
        self = addNode(tree, "IntLit(" + tree.value + ")");
    }

    public void caseApply(Apply tree) {
        self = addNode(tree, "Apply");
        child(self, tree.fun);
        child(self, tree.args);
    }

    private Node nodeFor(Tree tree) {
        Node bkpSelf = self;
        tree.apply(this);
        assert self != bkpSelf : "no node generated for " + tree;
        Node nodeForTree = self;
        self.color = WHITE;
        self = bkpSelf;
        return nodeForTree;
    }

    private void child(Node parent, Tree childTree, String name) {
        Node child = nodeFor(childTree);
        assert parent != null;
        assert child != null;
        addEdge(parent, child, name);
    }

    private void child(Node parent, Tree childTree) {
        child(parent, childTree, "");
    }

    private void child(Node parent, Tree[] childrenTrees) {
        for (int i = 0; i < childrenTrees.length; ++i)
            child(parent, childrenTrees[i]);
    }

    //////////////////////////////////////////////////////////////////////

    private HashMap/*<Tree, Node>*/ nodes = new HashMap();
    private LinkedList/*<Edge>*/ edges = new LinkedList();

    private int nodeCounter = 0;

    private Node addNode(Tree tree, String label) {
        assert !nodes.containsKey(tree);
        Node newNode = new Node(++nodeCounter, label);
        nodes.put(tree, newNode);
        return newNode;
    }

    private Edge addEdge(Node from, Node to, String label) {
        Edge newEdge = new Edge(from, to, label);
        edges.add(newEdge);
        return newEdge;
    }

    private static class Color {
        public final double hue, saturation, value;
        public Color(double hue, double saturation, double value) {
            this.hue = hue;
            this.saturation = saturation;
            this.value = value;
        }

        public String toString() {
            return "" + hue + "," + saturation + "," + value;
        }
    }

    private static Color WHITE = new Color(0.0, 0.0, 1.0);

    private static class Node {
        private final int n;
        public final String label;
        public Color color;
        public String toString() { return "n" + n; }
        public Node(int n, String label, Color color) {
            this.n = n;
            this.label = label;
            this.color = color;
        }
        public Node(int n, String label) {
            this(n, label, WHITE);
        }
    }

    private static class Edge {
        public final Node from, to;
        public final String label;
        public Edge(Node from, Node to, String label) {
            this.from = from;
            this.to = to;
            this.label = label;
        }
    }
}
