/*  miscc -- a compiler for misc
 *    Code generator
 *
 *    Michel Schinz
 *    $Id: Generator.java,v 1.4 2004/03/23 09:39:01 schinz Exp $
 */

package miscc;

import miscc.Tree.*;
import java.util.BitSet;

public class Generator {
    private final Scope globalScope;

    private final Gen genVisitor;
    private final GenLoad genLoadVisitor;
    private final GenCond genCondVisitor;
    public final Code code;

    public Generator(Scope globalScope) {
        this.globalScope = globalScope;
        this.genVisitor = new Gen();
        this.genLoadVisitor = new GenLoad();
        this.genCondVisitor = new GenCond();
        this.code = getCode();
    }

    protected Code getCode() { return new Code(); }

    public void gen(Tree tree) {
        genVisitor.generate(tree);
    }

    public void gen(Tree[] trees) {
        for (int i = 0; i < trees.length; ++i)
            gen(trees[i]);
    }

    public void genLoad(Tree tree, int targetReg) {
        assert code.usedRegisters().get(targetReg)
            : "invariant violated: target register R" + targetReg
            + " not allocated";
        genLoadVisitor.generate(tree, targetReg);
    }

    public void genCond(Tree tree, Code.Label targetLabel, boolean when) {
        genCondVisitor.generate(tree, targetLabel, when);
    }

    /**
     * Visitor generating code for statements, in a general sense,
     * that is all constructs which do not have a value.
     */
    private class Gen extends DefaultVisitor implements Visitor, RISC {
        public void generate(Tree tree) {
            tree.apply(this);
        }

        public void caseProgram(Program tree) {
            // initialise stack pointer
            code.emit(SYSCALL, SP, 0, SYS_GET_TOTAL_MEM_SIZE);
            // initialise garbage collector
            code.emit(DIVI, 1, SP, 4);
            code.emit(DIVI, 2, 1, 2);
            code.emit(ADDI, 3, 0, SP);
            code.emit(LSHI, 3, 3, 27);
            code.emit(ADD, 2, 2, 3);
            code.emit(SYSCALL, 1, 2, SYS_GC_INIT);

            // jump to main expression
            Code.Label startLabel = code.getLabel();
            code.emit(BEQ, ZERO, startLabel);

            // generate predefined functions
            for (Symbol sym = globalScope.first; sym != null; sym = sym.next) {
                sym.offset = code.pc();
                if (sym.name.equals("printInt")) {
                    code.emit(POP, 1, SP, WORD_SIZE);
                    code.emit(SYSCALL, 1, ZERO, SYS_IO_WR_INT);
                } else if (sym.name.equals("printChar")) {
                    code.emit(POP, 1, SP, WORD_SIZE);
                    code.emit(SYSCALL, 1, ZERO, SYS_IO_WR_CHR);
                } else if (sym.name.equals("readInt")) {
                    code.emit(SYSCALL, RES, ZERO, SYS_IO_RD_INT);
                } else if (sym.name.equals("readChar")) {
                    code.emit(SYSCALL, RES, ZERO, SYS_IO_RD_CHR);
                } else
                    throw new Error("unknown predefined function: " + sym.name);
                code.emit(RET, LNK);
            }

            // generete function definitions
            gen(tree.defs);

            // generate main expression
            code.anchorLabel(startLabel);
            gen(tree.body);

            // exit
            code.emit(RET, ZERO);
        }

        public void caseFunDef(FunDef tree) {
            tree.sym.offset = code.pc();
            gen(tree.args);
            code.incFrameSize(WORD_SIZE);
            code.emit(PSH, LNK, SP, WORD_SIZE);
            int resReg = code.getRegister(RES);
            genLoad(tree.body, resReg);
            int popSize = WORD_SIZE * (1 + tree.args.length);
            code.emit(POP, LNK, SP, popSize);
            code.decFrameSize(popSize);
            assert code.getFrameSize() == 0
                : "non-empty frame (size " + code.getFrameSize() + ")";
            code.emit(RET, LNK);
            code.freeRegister(resReg);
        }

        public void caseFormal(Formal tree) {
            code.incFrameSize(WORD_SIZE);
            tree.sym.offset = code.getFrameSize();
        }

        public void caseBasicType(BasicType tree) {
            throw new Error("no code to generate for " + tree);
        }

        public void caseListType(ListType tree) {
            throw new Error("no code to generate for " + tree);
        }

        public void caseFunType(FunType tree) {
            throw new Error("no code to generate for " + tree);
        }

        public void caseVarDef(VarDef tree) {
            int tmpReg = code.getRegister();
            genLoad(tree.init, tmpReg);
            code.incFrameSize(WORD_SIZE);
            code.emit(PSH, tmpReg, SP, WORD_SIZE);
            tree.var.sym.offset = code.getFrameSize();
            code.freeRegister(tmpReg);
        }

        public void caseWhile(While tree) {
            Code.Label startLabel = code.getLabel();
            code.anchorLabel(startLabel);
            Code.Label afterLabel = code.getLabel();
            genCond(tree.cond, afterLabel, false);
            gen(tree.body);
            code.emit(BEQ, 0, startLabel);
            code.anchorLabel(afterLabel);
        }

        public void caseExec(Exec tree) {
            gen(tree.expr);
        }

        public void caseDefault(Tree tree) {
            int tmpReg = code.getRegister();
            genLoad(tree, tmpReg);
            code.freeRegister(tmpReg);
        }
    }

    /**
     * Visitor generating code for expressions, that is constructs
     * which produce a value.
     */
    private class GenLoad extends DefaultVisitor implements Visitor, RISC {
        private int targetReg;

        public void generate(Tree tree, int targetReg) {
            int bkpTargetReg = this.targetReg;
            this.targetReg = targetReg;
            tree.apply(this);
            this.targetReg = bkpTargetReg;
        }

        public void caseBody(Body tree) {
            int beginFrameSize = code.getFrameSize();
            code.freeRegister(targetReg);
            gen(tree.stats);
            code.getRegister(targetReg);
            genLoad(tree.expr, targetReg);
            int endFrameSize = code.getFrameSize();
            if (endFrameSize > beginFrameSize) {
                code.emit(ADDI, SP, SP, endFrameSize - beginFrameSize);
                code.decFrameSize(endFrameSize - beginFrameSize);
            }
        }

        public void caseIf(If tree) {
            Code.Label elseLabel = code.getLabel();
            code.freeRegister(targetReg);
            genCond(tree.cond, elseLabel, false);
            code.getRegister(targetReg);
            genLoad(tree.thenp, targetReg);
            Code.Label afterLabel = code.getLabel();
            code.emit(BEQ, 0, afterLabel);
            code.anchorLabel(elseLabel);
            genLoad(tree.elsep, targetReg);
            code.anchorLabel(afterLabel);
        }

        public void caseAssign(Assign tree) {
            genLoad(tree.expr, targetReg);
            int realOffset = code.getFrameSize() - tree.sym.offset;
            code.emit(STW, targetReg, SP, realOffset);
        }

        public void caseOperation(Operation tree) {
            switch (tree.op) {
            case Tree.ADD: case Tree.SUB:
            case Tree.MUL: case Tree.DIV: case Tree.MOD: {
                genLoad(tree.left, targetReg);
                int rightReg = code.getRegister();
                genLoad(tree.right, rightReg);
                code.emit(opcode(tree.op), targetReg, targetReg, rightReg);
                code.freeRegister(rightReg);
            } break;

            case Tree.CONS: {
                int tmpReg = code.getRegister();
                code.freeRegister(targetReg);
                genLoad(tree.left, tmpReg);
                code.getRegister(targetReg);
                code.emit(ADDI, targetReg, 0, 2 * WORD_SIZE);
                code.emit(SYSCALL, targetReg, targetReg, SYS_GC_ALLOC);
                code.emit(STW, tmpReg, targetReg, 0);
                genLoad(tree.right, tmpReg);
                code.emit(STW, tmpReg, targetReg, WORD_SIZE);
                code.freeRegister(tmpReg);
            } break;

            default:
                caseDefault(tree);
            }
        }

        public void caseListop(Listop tree) {
            switch (tree.op) {
            case Tree.HEAD:
            case Tree.TAIL: {
                int offset = WORD_SIZE * (tree.op == Tree.HEAD ? 0 : 1);
                genLoad(tree.args[0], targetReg);
                Code.Label okLabel = code.getLabel();
                code.emit(BNE, targetReg, okLabel);
                code.emit(ADDI, 1, ZERO, 1 + offset);
                code.emit(SYSCALL, 1, ZERO, SYS_EXIT);
                code.anchorLabel(okLabel);
                code.emit(LDW, targetReg, targetReg, offset);
            } break;

            case Tree.ISEMPTY:
                caseDefault(tree);
                break;

            default:
                throw new Error("unknown list operator " + tree.op);
            }
        }

        public void caseIdent(Ident tree) {
            if (tree.sym.isVariable) {
                int realOffset = code.getFrameSize() - tree.sym.offset;
                code.emit(LDW, targetReg, SP, realOffset);
            } else
                emitLoadConstant(targetReg, tree.sym.offset);
        }

        public void caseUnitLit(UnitLit tree) {
            // nothing to do
        }

        public void caseNilLit(NilLit tree) {
            code.emit(ADDI, targetReg, 0, 0);
        }

        public void caseIntLit(IntLit tree) {
            emitLoadConstant(targetReg, tree.value);
        }

        public void caseApply(Apply tree) {
            int funAddr = -1;
            int funReg = -1;
            if ((tree.fun instanceof Ident) && !((Ident)tree.fun).sym.isVariable)
                funAddr = ((Ident)tree.fun).sym.offset;

            // save registers
            code.freeRegister(targetReg);
            BitSet usedRegs = code.usedRegisters();
            for (int r = usedRegs.nextSetBit(0);
                 r >= 0;
                 r = usedRegs.nextSetBit(r+1)) {
                code.incFrameSize(WORD_SIZE);
                code.emit(PSH, r, SP, WORD_SIZE);
                code.freeRegister(r);
            }

            if (funAddr == -1) {
                funReg = code.getRegister();
                genLoad(tree.fun, funReg);
            }

            // load arguments
            for (int i = 0; i < tree.args.length; ++i) {
                int tmpReg = code.getRegister();
                genLoad(tree.args[i], tmpReg);
                code.incFrameSize(WORD_SIZE);
                code.emit(PSH, tmpReg, SP, WORD_SIZE);
                code.freeRegister(tmpReg);
            }

            if (funAddr != -1)
                code.emit(JSR, funAddr / 4);
            else {
                code.emit(ORIU, LNK, 0, code.pc() + WORD_SIZE * 2);
                code.emit(RET, funReg);
                code.freeRegister(funReg);
            }

            code.decFrameSize(tree.args.length * WORD_SIZE);
            code.getRegister(targetReg);
            if (targetReg != RES)
                code.emit(ADDI, targetReg, RES, 0);

            // restore registers
            for (int r = usedRegs.nextSetBit(0);
                 r >= 0;
                 r = usedRegs.nextSetBit(r+1)) {
                code.getRegister(r);
                code.emit(POP, r, SP, WORD_SIZE);
                code.decFrameSize(WORD_SIZE);
            }
        }

        public void caseDefault(Tree tree) {
            genLoad(new If(tree.pos,
                           tree,
                           new IntLit(tree.pos, 1),
                           new IntLit(tree.pos, 0)),
                    targetReg);
        }

        private void emitLoadConstant(int reg, int value) {
            int highBits = value >> 16;
            int lowBits = value & 0xFFFF;
            if (highBits != 0) {
                code.emit(ORIU, reg, 0, highBits);
                code.emit(LSHI, reg, reg, 16);
                code.emit(ORIU, reg, reg, lowBits);
            } else
                code.emit(ORIU, reg, 0, lowBits);
        }

        private int opcode(int op) {
            switch (op) {
            case Tree.ADD: return ADD;
            case Tree.SUB: return SUB;
            case Tree.MUL: return MUL;
            case Tree.DIV: return DIV;
            case Tree.MOD: return MOD;
            default: throw new Error("unknown opcode: " + op);
            }
        }
    }

    /**
     * Visitor generating code for conditions, that is constructs
     * which jump somewhere on the result of a test.
     */
    private class GenCond extends DefaultVisitor implements Visitor, RISC {
        private Code.Label targetLabel;
        private boolean when;

        public void generate(Tree tree, Code.Label targetLabel, boolean when) {
            Code.Label bkpTargetLabel = this.targetLabel;
            this.targetLabel = targetLabel;
            boolean bkpWhen = this.when;
            this.when = when;
            tree.apply(this);
            this.targetLabel = bkpTargetLabel;
            this.when = bkpWhen;
        }

        public void caseOperation(Operation tree) {
            switch (tree.op) {
            case Tree.EQ: case Tree.NE: case Tree.LT:
            case Tree.LE: case Tree.GT: case Tree.GE:
                int leftReg = code.getRegister();
                genLoad(tree.left, leftReg);
                int rightReg = code.getRegister();
                genLoad(tree.right, rightReg);
                code.emit(CMP, leftReg, leftReg, rightReg);
                code.freeRegister(rightReg);
                code.emit(opcode(tree.op, when), leftReg, targetLabel);
                code.freeRegister(leftReg);
                break;

            case Tree.AND:
                if (when) {
                    Code.Label afterLabel = code.getLabel();
                    genCond(tree.left, afterLabel, false);
                    genCond(tree.right, targetLabel, true);
                    code.anchorLabel(afterLabel);
                } else {
                    genCond(tree.left, targetLabel, false);
                    genCond(tree.right, targetLabel, false);
                }
                break;

            case Tree.NOT:
                genCond(tree.left, targetLabel, !when);
                break;

            default:
                caseDefault(tree);
            }
        }

        public void caseListop(Listop tree) {
            if (tree.op == Tree.ISEMPTY)
                genCond(tree.args[0], targetLabel, !when);
            else
                caseDefault(tree);
        }

        public void caseDefault(Tree tree) {
            int tmpReg = code.getRegister();
            genLoad(tree, tmpReg);
            code.emit((when ? BNE : BEQ), tmpReg, targetLabel);
            code.freeRegister(tmpReg);
        }

        private int opcode(int op, boolean negated) {
            switch (op) {
            case Tree.EQ : return (negated ? BEQ : BNE);
            case Tree.NE : return (negated ? BNE : BEQ);
            case Tree.LT : return (negated ? BLT : BGE);
            case Tree.LE : return (negated ? BLE : BGT);
            case Tree.GT : return (negated ? BGT : BLE);
            case Tree.GE : return (negated ? BGE : BLT);
            default      : throw new Error("invalid operator: " + op);
            }
        }
    }
}
