/*  miscc -- a compiler for misc
 *    Representation of the abstract syntax
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Printer.java,v 1.1 2004/03/01 05:34:16 schinz Exp $
 */

package miscc;

import miscc.Tree.*;


/** This class implements a pretty printer for abstract syntax
 *  trees in misc.
 */
class Printer implements Visitor {

    /** Indentation string.
     */
    String step = "  ";

    /** indentation level.
     */
    int level = 0;

    /** Print a tree node.
     */
    Printer print(Tree tree) {
        tree.apply(this);
        return this;
    }

    /** Print abstract syntax trees separated by commas.
     */
    Printer print(Tree[] trees) {
        for(int i = 0; i < trees.length; i++) {
            if (i > 0) print(", ");
            print(trees[i]);
        }
        return this;
    }

    /** Print a number.
     */
    Printer print(int value) {
        return print(String.valueOf(value));
    }

    /** Print a string.
     */
    Printer print(String value) {
        System.out.print(value);
        return this;
    }

    /** Print an end-of-line character and indent the following
     *  line appropriately.
     */
    Printer println() {
        System.out.println();
        for (int i = 0; i < level; i++)
            System.out.print(step);
        return this;
    }

    /** Increment the indentation level.
     */
    Printer indent() {
        level++;
        return this;
    }

    /** Decrement the indentation level.
     */
    Printer undent() {
        level--;
        return this;
    }

    //// VISITOR IMPLEMENTATION ////

    public void caseBad(Bad tree) {
        print("<bad>");
    }

    public void caseProgram(Program tree) {
        for (int i = 0; i < tree.defs.length; i++)
            print(tree.defs[i]).print(";").println();
        if (tree.body.stats.length > 0)
            for (int i = 0; i < tree.body.stats.length; i++)
                print(tree.body.stats[i]).print(";").println();
        print(tree.body.expr).println();
    }

    public void caseBody(Body tree) {
        if (tree.stats.length > 0) {
            print("{").indent().println();
            for (int i = 0; i < tree.stats.length; i++)
                print(tree.stats[i]).print(";").println();
            print(tree.expr).undent().println().print("}");
        } else
            print(tree.expr);
    }

    public void caseFunDef(FunDef tree) {
        print("def ").print(tree.name).print("(").print(tree.args).print(")");
        print(": ").print(tree.restype).print(" = ").indent();
        println().print(tree.body).undent();
    }

    public void caseFormal(Formal tree) {
        print(tree.name).print(": ").print(tree.vartype);
    }

    public void caseBasicType(BasicType tree) {
        switch (tree.tag) {
            case Tree.UNIT:
                print("Unit");
                break;
            case Tree.ANY:
                print("Any");
                break;
            case Tree.INT:
                print("Int");
                break;
            default:
                throw new Error("unknown type tag.");
        }
    }

    public void caseListType(ListType tree) {
        print("List[").print(tree.elemtype).print("]");
    }

    public void caseFunType(FunType tree) {
        print("(").print(tree.argtypes).print(") => ").print(tree.restype);
    }

    public void caseVarDef(VarDef tree) {
        print("var ").print(tree.var).print(" = ").indent();
        print(tree.init).undent();
    }

    public void caseWhile(While tree) {
        print("while (").indent().print(tree.cond).undent().print(")");
        indent().println().print(tree.body).undent();
    }

    public void caseExec(Exec tree) {
        print(tree.expr);
    }

    public void caseIf(If tree) {
        print("if (").indent().print(tree.cond).undent().print(")");
        indent().println().print(tree.thenp);
        undent().println().print("else");
        indent().println().print(tree.elsep).undent();
    }

    public void caseAssign(Assign tree) {
        print(tree.name + " = ").print(tree.expr);
    }

    public void caseOperation(Operation tree) {
        if (tree.right != null) {
            print("(").print(tree.left);
            switch (tree.op) {
            case Tree.ADD:
                print(" + "); break;
            case Tree.SUB:
                print(" - "); break;
            case Tree.MUL:
                print(" * "); break;
            case Tree.DIV:
                print(" / "); break;
            case Tree.MOD:
                print(" % "); break;
            case Tree.EQ:
                print(" == "); break;
            case Tree.NE:
                print(" != "); break;
            case Tree.LT:
                print(" < "); break;
            case Tree.LE:
                print(" <= "); break;
            case Tree.GT:
                print(" > "); break;
            case Tree.GE:
                print(" >= "); break;
            case Tree.AND:
                print(" & "); break;
            case Tree.CONS:
                print(" :: "); break;
            default:
                throw new Error();
            }
            print(tree.right).print(")");
        } else {
            switch (tree.op) {
            case Tree.NOT:
                print("(!").print(tree.left).print(")"); break;
            default:
                throw new Error();
            }
        }
    }

    public void caseListop(Listop tree) {
        switch (tree.op) {
            case Tree.HEAD:
                print("head(").print(tree.args).print(")");
                break;
            case Tree.TAIL:
                print("tail(").print(tree.args).print(")");
                break;
            case Tree.ISEMPTY:
                print("isEmpty(").print(tree.args).print(")");
                break;
            default:
                throw new Error();
        }
    }

    public void caseIdent(Ident tree) {
        print(tree.name);
    }

    public void caseUnitLit(UnitLit tree) {
        print("()");
    }

    public void caseNilLit(NilLit tree) {
        print("List()");
    }

    public void caseIntLit(IntLit tree) {
        print("" + tree.value);
    }

    public void caseApply(Apply tree) {
        print(tree.fun).print("(").print(tree.args).print(")");
    }
}
