/*  miscc -- a compiler for misc
 *    Scanner
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Scanner.java,v 1.1 2004/03/01 05:34:13 schinz Exp $
 */

package miscc;

import java.io.InputStream;
import java.io.IOException;
import java.util.Map;
import java.util.HashMap;


/** This class implements the scanner of the compiler.
 */
class Scanner implements Tokens {

    /** This character represents the end of the input.
     */
    private static final char EOF_CH = (char)-1;

    /** A table that associates a keyword with its corresponding
     *  token class.
     */
    private static final Map keywords = new HashMap();
    static {
        keywords.put("def"    , new Integer(DEF));
        keywords.put("var"    , new Integer(VAR));
        keywords.put("while"  , new Integer(WHILE));
        keywords.put("if"     , new Integer(IF));
        keywords.put("else"   , new Integer(ELSE));
        keywords.put("Unit"   , new Integer(UNIT));
        keywords.put("Any"    , new Integer(ANY));
        keywords.put("Int"    , new Integer(INT));
        keywords.put("List"   , new Integer(LIST));
        keywords.put("false"  , new Integer(FALSE));
        keywords.put("true"   , new Integer(TRUE));
        keywords.put("head"   , new Integer(HEAD));
        keywords.put("tail"   , new Integer(TAIL));
        keywords.put("isEmpty", new Integer(IS_EMPTY));
    }

    /** The current token class.
     */
    int token;

    /** The position of the first character of the current token.
     */
    int start;

    /** The string representation of the current token. This variable
     *  is only set if the current token class has several possible
     *  textual representations.
     */
    String chars;

    /** A buffer for constructing string representations of tokens.
     */
    private StringBuffer buf;

    /** The current character.
     */
    private char ch;

    /** The line and the column of the current token.
     */
    private int line = 1;
    private int column = 0;

    /** The input stream providing the input characters.
     */
    private final InputStream in;


    Scanner(InputStream in) {
        this.in = in;
        this.buf = new StringBuffer();
        nextCh();
        nextToken();
    }

    /** close the file input stream
     */
    void close() throws IOException {
        in.close();
    }

    /** This method reads the next token and stores the token class
     *  in variable 'token'. If the token representation is not unique
     *  it will also leave a textual representation in variable 'chars'.
     *  Whitespaces and comments are skipped by this method.
     */
    void nextToken() {
        // skip comments and whitespace characters
        while (true) {
            while (Character.isWhitespace(ch))
                nextCh();
            if (ch == '/') {
                start = Position.encode(line, column);
                nextCh();
                if (ch == '/') {
                    while (ch != '\n' && ch != EOF_CH)
                        nextCh();
                } else {
                    token = DIV;
                    return;
                }
            } else
                break;
        }
        // initialize the position of the current token
        start = Position.encode(line, column);
        // read the current token
        token = readToken();
    }

    /** Read the next token, store its representation (if its not
     *  unique) in variable 'chars' and return the token class.
     */
    private int readToken() {
        switch (ch) {
            // end of file
            case EOF_CH:
                return EOF;
            // identifier or keyword
            case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
            case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
            case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
            case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
            case 'Y': case 'Z':
            case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
            case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
            case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
            case 's': case 't': case 'u': case 'v': case 'w': case 'x':
            case 'y': case 'z':
                buf.setLength(0);
                do {
                    buf.append(ch);
                    nextCh();
                } while (Character.isLetterOrDigit(ch) || ch == '_');
                chars = buf.toString();
                Object object = keywords.get(chars);
                return object != null ? ((Integer)object).intValue() : IDENT;
            // zero
            case '0':
                chars = "0";
                nextCh();
                return NUMBER;
            // non-zero numbers
            case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
                buf.setLength(0);
                do {
                    buf.append(ch);
                    nextCh();
                } while (Character.isDigit(ch));
                chars = buf.toString();
                return NUMBER;
            // strings
            case '"':
                nextCh();
                buf.setLength(0);
                while (ch != '"') {
                    if (ch == '\n' || ch == EOF_CH)
                        Report.fail(start, "string not properly terminated");
                    buf.append(ch);
                    nextCh();
                }
                nextCh();
                chars = buf.toString();
                return STRING;
            // operators and special characters
            case '(':
                nextCh();
                return LPAREN;
            case ')':
                nextCh();
                return RPAREN;
            case '{':
                nextCh();
                return LBRACE;
            case '}':
                nextCh();
                return RBRACE;
            case '[':
                nextCh();
                return LBRACKET;
            case ']':
                nextCh();
                return RBRACKET;
            case ',':
                nextCh();
                return COMMA;
            case ';':
                nextCh();
                return SEMICOLON;
            case ':':
                nextCh();
                if (ch == ':') {
                    nextCh();
                    return CONS;
                } else
                    return COLON;
            case '=':
                nextCh();
                if (ch == '=') {
                    nextCh();
                    return EQ;
                } else if (ch == '>') {
                    nextCh();
                    return ARROW;
                } else
                    return EQUAL;
            case '!':
                nextCh();
                if (ch == '=') {
                    nextCh();
                    return NE;
                } else
                    return NOT;
            case '<':
                nextCh();
                if (ch == '=') {
                    nextCh();
                    return LE;
                } else
                    return LT;
            case '>':
                nextCh();
                if (ch == '=') {
                    nextCh();
                    return GE;
                } else
                    return GT;
            case '+':
                nextCh();
                return ADD;
            case '-':
                nextCh();
                return SUB;
            case '*':
                nextCh();
                return MUL;
            case '/':
                nextCh();
                return DIV;
            case '%':
                nextCh();
                return MOD;
            case '&':
                nextCh();
                return AND;
            case '|':
                nextCh();
                return OR;
            default:
                Report.fail(start, "invalid character: '" + ch + "'");
                return BAD;
        }
    }

    /** Return a textual representation of the current token.
     */
    String representation() {
        String representation = tokenClass(token);
        if (token == NUMBER || token == IDENT || token == STRING)
            representation += "(" + chars + ")";
        return representation;
    }

    /** Return the textual representation of the given token class.
     */
    static String tokenClass(int token) {
        switch (token) {
            case EOF:
                return "<eof>";
            case BAD:
                return "<bad>";
            case IDENT:
                return "ident";
            case NUMBER:
                return "number";
            case STRING:
                return "string";
            case LPAREN:
                return "(";
            case RPAREN:
                return ")";
            case LBRACE:
                return "{";
            case RBRACE:
                return "}";
            case LBRACKET:
                return "[";
            case RBRACKET:
                return "]";
            case COMMA:
                return ",";
            case SEMICOLON:
                return ";";
            case COLON:
                return ":";
            case CONS:
                return "::";
            case EQUAL:
                return "=";
            case EQ:
                return "==";
            case NOT:
                return "!";
            case NE:
                return "!=";
            case LT:
                return "<";
            case LE:
                return "<=";
            case GT:
                return ">";
            case GE:
                return ">=";
            case ARROW:
                return "=>";
            case ADD:
                return "+";
            case SUB:
                return "-";
            case MUL:
                return "*";
            case DIV:
                return "/";
            case MOD:
                return "%";
            case AND:
                return "&";
            case OR:
                return "|";
            case DEF:
                return "def";
            case VAR:
                return "var";
            case WHILE:
                return "while";
            case IF:
                return "if";
            case ELSE:
                return "else";
            case UNIT:
                return "Unit";
            case ANY:
                return "Any";
            case INT:
                return "Int";
            case LIST:
                return "List";
            case FALSE:
                return "false";
            case TRUE:
                return "true";
            case HEAD:
                return "head";
            case TAIL:
                return "tail";
            case IS_EMPTY:
                return "isEmpty";
            default:
                throw new Error("unknown token class " + token);
        }
    }

    /** Puts the next character into 'ch' and updates the current position.
     */
    private void nextCh() {
        switch (ch) {
            case EOF_CH:
                return;
            case '\n':
                column = 1;
                line++;
                break;
            default:
                column++;
        }
        try {
            ch = (char)in.read();
            oldch = ((oldch == '\r') && (ch == '\n')) ? (char)in.read() : ch;
            ch = (oldch == '\r') ? '\n' : oldch;
        } catch (IOException e) {
            Report.fail(Position.encode(line, column), e.getMessage());
        }
    }

    private char oldch;
}
