/*  miscc -- a compiler for misc
 *    Representation of the abstract syntax
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Tree.java,v 1.1 2004/03/01 05:34:10 schinz Exp $
 */

package miscc;


/** Class Tree defines the type of all abstract syntax tree nodes.
 */
abstract class Tree {

    /** The source code position of this tree node.
     */
    int pos;

    /** The type of the tree.
     */
    Type type;


    /** Constructor for tree nodes.
     */
    Tree(int pos) {
        this.pos = pos;
    }

    /** Applies the given visitor object to this tree node.
     */
    abstract void apply(Visitor v);

    /** This node represents an erroneous syntactical construct.
     */
    static class Bad extends Tree {
        Bad(int pos) {
            super(pos);
        }

        void apply(Visitor v) {
            v.caseBad(this);
        }
    }

    /** Program { D } B
     */
    static class Program extends Tree {
        Tree[] defs;
        Body body;

        Program(int pos, Tree[] defs, Body body) {
            super(pos);
            this.defs = defs;
            this.body = body;
        }

        void apply(Visitor v) {
            v.caseProgram(this);
        }
    }

    /** Body { S } E
     */
    static class Body extends Tree {
        Tree[] stats;
        Tree expr;

        Body(int pos, Tree[] stats, Tree expr) {
            super(pos);
            this.stats = stats;
            this.expr = expr;
        }

        void apply(Visitor v) {
            v.caseBody(this);
        }
    }

    /** FunDef ident { F } T E
     */
    static class FunDef extends Tree {
        String name;
        Tree[] args;
        Tree restype;
        Tree body;
        Symbol sym;

        FunDef(int pos, String name, Tree[] args, Tree restype, Tree body) {
            super(pos);
            this.name = name;
            this.args = args;
            this.restype = restype;
            this.body = body;
        }

        void apply(Visitor v) {
            v.caseFunDef(this);
        }
    }

    /** Formal ident T
     */
    static class Formal extends Tree {
        String name;
        Tree vartype;
        Symbol sym;

        Formal(int pos, String name, Tree vartype) {
            super(pos);
            this.name = name;
            this.vartype = vartype;
        }

        void apply(Visitor v) {
            v.caseFormal(this);
        }
    }

    /** BasicType U
     */
    static class BasicType extends Tree {
        int tag;

        BasicType(int pos, int tag) {
            super(pos);
            this.tag = tag;
        }

        void apply(Visitor v) {
            v.caseBasicType(this);
        }
    }

    /** Basic type tags.
     */
    static final int UNIT = 0;
    static final int ANY  = 1;
    static final int INT  = 2;

    /** ListType T
     */
    static class ListType extends Tree {
        Tree elemtype;

        ListType(int pos, Tree elemtype) {
            super(pos);
            this.elemtype = elemtype;
        }

        void apply(Visitor v) {
            v.caseListType(this);
        }
    }

    /** FunType { T } T
     */
    static class FunType extends Tree {
        Tree[] argtypes;
        Tree restype;

        FunType(int pos, Tree[] argtypes, Tree restype) {
            super(pos);
            this.argtypes = argtypes;
            this.restype = restype;
        }

        void apply(Visitor v) {
            v.caseFunType(this);
        }
    }

    /** VarDef ident T E
     */
    static class VarDef extends Tree {
        Formal var;
        Tree init;

        VarDef(int pos, Formal var, Tree init) {
            super(pos);
            this.var = var;
            this.init = init;
        }

        void apply(Visitor v) {
            v.caseVarDef(this);
        }
    }

    /** While E S
     */
    static class While extends Tree {
        Tree cond;
        Tree body;

        While(int pos, Tree cond, Tree body) {
            super(pos);
            this.cond = cond;
            this.body = body;
        }

        void apply(Visitor v) {
            v.caseWhile(this);
        }
    }

    /** Exec E
     */
    static class Exec extends Tree {
        Tree expr;

        Exec(int pos, Tree expr) {
            super(pos);
            this.expr = expr;
        }

        void apply(Visitor v) {
            v.caseExec(this);
        }
    }

    /** If E E E
     */
    static class If extends Tree {
        Tree cond;
        Tree thenp;
        Tree elsep;

        If(int pos, Tree cond, Tree thenp, Tree elsep) {
            super(pos);
            this.cond = cond;
            this.thenp = thenp;
            this.elsep = elsep;
        }

        void apply(Visitor v) {
            v.caseIf(this);
        }
    }

    /** Assign ident E
     */
    static class Assign extends Tree {
        String name;
        Tree expr;
        Symbol sym;

        Assign(int pos, String name, Tree expr) {
            super(pos);
            this.name = name;
            this.expr = expr;
        }

        void apply(Visitor v) {
            v.caseAssign(this);
        }
    }

    /** Operation O E [ E ]
     */
    static class Operation extends Tree {
        int op;
        Tree left;
        Tree right;

        Operation(int pos, int op, Tree left, Tree right) {
            super(pos);
            this.op = op;
            this.left = left;
            this.right = right;
        }

        Operation(int pos, int op, Tree expr) {
            this(pos, op, expr, null);
        }

        void apply(Visitor v) {
            v.caseOperation(this);
        }
    }

    static final int ADD = 0;
    static final int SUB = 1;
    static final int MUL = 2;
    static final int DIV = 3;
    static final int MOD = 4;

    static final int EQ = 5;
    static final int NE = 6;
    static final int LT = 7;
    static final int LE = 8;
    static final int GT = 9;
    static final int GE = 10;

    static final int NOT = 11;
    static final int AND = 12;

    static final int CONS = 13;

    /** Listop L E
     */
    static class Listop extends Tree {
        int op;
        Tree[] args;

        Listop(int pos, int op, Tree[] args) {
            super(pos);
            this.op = op;
            this.args = args;
        }

        void apply(Visitor v) {
            v.caseListop(this);
        }
    }

    static final int HEAD = 0;
    static final int TAIL = 1;
    static final int ISEMPTY = 2;

    /** Ident ident
     */
    static class Ident extends Tree {
        String name;
        Symbol sym;

        Ident(int pos, String name) {
            super(pos);
            this.name = name;
        }

        void apply(Visitor v) {
            v.caseIdent(this);
        }
    }

    /** UnitLit
     */
    static class UnitLit extends Tree {
        UnitLit(int pos) {
            super(pos);
        }

        void apply(Visitor v) {
            v.caseUnitLit(this);
        }
    }

    /** NilLit
     */
    static class NilLit extends Tree {
        NilLit(int pos) {
            super(pos);
        }

        void apply(Visitor v) {
            v.caseNilLit(this);
        }
    }

    /** IntLit
     */
    static class IntLit extends Tree {
        int value;

        IntLit(int pos, int value) {
            super(pos);
            this.value = value;
        }

        void apply(Visitor v) {
            v.caseIntLit(this);
        }
    }

    /** Apply E { E }
     */
    static class Apply extends Tree {
        Tree fun;
        Tree[] args;

        Apply(int pos, Tree fun, Tree[] args) {
            super(pos);
            this.fun = fun;
            this.args = args;
        }

        void apply(Visitor v) {
            v.caseApply(this);
        }
    }
}
